package FindEscape;

public class FindEscape {
    // escEnv maps Symbol to Escape
    private Symbol.Table escEnv;

    private void traverseVar(int depth, Absyn.Var v) {
	if (v instanceof Absyn.SimpleVar)
	    traverseVar(depth,(Absyn.SimpleVar)v);
	else if (v instanceof Absyn.FieldVar)
	    traverseVar(depth,(Absyn.FieldVar)v);
	else if (v instanceof Absyn.SubscriptVar)
	    traverseVar(depth,(Absyn.SubscriptVar)v);

	throw new Error("traverseVar");
    }

    public void traverseExp(int depth, Absyn.Exp e) {}

    private void traverseDec(int depth, Absyn.Dec d) {
	if (d instanceof Absyn.FunctionDec)
	    traverseDec(depth,(Absyn.FunctionDec)d);
	else if (d instanceof Absyn.VarDec)
	    traverseDec(depth,(Absyn.VarDec)d);
	else if (d instanceof Absyn.TypeDec)
	    traverseDec(depth,(Absyn.TypeDec)d);

	throw new Error("traverseDec");
    }

    public FindEscape(Absyn.Exp e) {
	escEnv = new Symbol.Table();
	traverseExp(0,e);
    }

    private void traverseVar(int depth,Absyn.SimpleVar v) {
	Escape ent=(Escape)escEnv.get(v.name);

	if (ent!=null && ent.depth<depth) ent.setEscape();
    }

    private void traverseVar(int depth,Absyn.FieldVar v) {
	traverseVar(depth,v.var);
    }

    private void traverseVar(int depth,Absyn.SubscriptVar v) {
	traverseVar(depth,v.var);
	traverseExp(depth,v.index);
    }

    private void traverseDec(int depth,Absyn.FunctionDec d) {
	depth++;

	escEnv.beginScope();

	for (;d!=null;d=d.next) {
	    processParams(depth,d.params);
	    traverseExp(depth,d.body);
	}

	escEnv.endScope();
    }

    private void traverseDec(int depth,Absyn.TypeDec d) {
	return;
    }

    private void traverseDec(int depth,Absyn.VarDec d) {
	escEnv.put(d.name,new VarEscape(depth,d));
	traverseExp(depth,d.init);
    }

    private void processParams(int depth,Absyn.FieldList p) {
	for (;p!=null;p=p.tail)
	    escEnv.put(p.name,new FormalEscape(depth,p));
    }
}
