package Translate;

public class Translate {
    private Temp.Label error = new Temp.Label("error");
    private Temp.Label malloc = new Temp.Label("malloc");

    private Frag frags = null;

    public Frag getResult() {
	return frags;
    }

    public void procEntryExit(Level level, Exp body) {
	Tree.Stm wrappedBody;

	if (body instanceof Ex)
	    wrappedBody = level.frame.procEntryExit1(body.unEx());
	else
	    wrappedBody = level.frame.procEntryExit1(body.unNx());

	frags = new ProcFrag(wrappedBody,level.frame,frags);
    }

    public void printFrags() {
	Frag cursor;

	for (cursor = frags; cursor != null; cursor = cursor.next) {
	    cursor.print();
	    System.out.println();
	}
    }

    public static void print(Exp code) {
	if (code instanceof Nx)
	    (new Tree.Print(System.out)).prStm(code.unNx());
	else
	    (new Tree.Print(System.out)).prExp(code.unEx());
    }

    public Exp SimpleVar(Access a, Level l) {
	Level d = a.home;

	Tree.Exp base = new Tree.TEMP(l.frame.FP());

	for (; l != d; l = l.parent) base = (l.frame.formals.head).exp(base);

	return new Ex(a.acc.exp(base));
    }

    public Exp FieldVar(Exp record, int offset, Level l) {
	offset *= l.frame.wordSize();

	return new Ex(new Tree.MEM(new Tree.BINOP(Tree.BINOP.PLUS,
	  record.unEx(),new Tree.CONST(offset))));
    }

    public Exp SubscriptVar(Exp array, Exp index, Level l) {
	return new Ex(new Tree.MEM(new Tree.BINOP(Tree.BINOP.PLUS,
	  new Tree.MEM(array.unEx()),new Tree.BINOP(Tree.BINOP.MUL,
	    index.unEx(),new Tree.CONST(l.frame.wordSize())))));
    }

    public Exp VarExp(Exp lvalue) {
	return lvalue;
    }

    public Exp NilExp() {
	return new Ex(new Tree.CONST(0));
    }

    public Exp IntExp(int value) {
	return new Ex(new Tree.CONST(value));
    }

    public Exp StringExp(String value, Level level) {
	Temp.Label l = new Temp.Label();

	frags = new DataFrag(l,level.frame.string(l,value),frags);

	return new Ex(new Tree.NAME(l));
    }

    public Exp CallExp(Temp.Label name, ExpList pars,
		       Level callee, Level caller, boolean noResult){
	Tree.Exp FP;
	Tree.ExpList newPars, cursor;

	if (callee == null) FP = new Tree.CONST(0);
	else {
	    FP = new Tree.TEMP(caller.frame.FP());

	    if (caller != callee.parent)
		if (caller.parent == callee.parent)
		    FP = (caller.frame.formals.head).exp(FP);
		else {
		    for (; caller != callee; caller = caller.parent)
			FP = (caller.frame.formals.head).exp(FP);

		    FP = (caller.frame.formals.head).exp(FP);
		}
	}

	newPars = cursor = new Tree.ExpList(FP,null);
	while (pars != null) {
	    cursor.tail = new Tree.ExpList(pars.head.unEx(),null);
	    cursor = cursor.tail;
	    pars = pars.tail;
	}

	if (noResult)
	    return new Nx(new Tree.EXP(new Tree.CALL(new Tree.NAME(name),
					      newPars)));
	else
	    return new Ex(new Tree.CALL(new Tree.NAME(name),newPars));
    }

    public Exp OpExp(Exp left, int oper, Exp right) {
	Tree.Exp l = left.unEx();
	Tree.Exp r = right.unEx();
	Exp lt = new Ex(l);
	Exp rt = new Ex(r);

	switch (oper) {
	case Absyn.OpExp.PLUS:
	    return new Ex(new Tree.BINOP(Tree.BINOP.PLUS,l,r));
	case Absyn.OpExp.MINUS:
	    return new Ex(new Tree.BINOP(Tree.BINOP.MINUS,l,r));
	case Absyn.OpExp.MUL:
	    return new Ex(new Tree.BINOP(Tree.BINOP.MUL,l,r));
	case Absyn.OpExp.DIV:
	    return new Ex(new Tree.BINOP(Tree.BINOP.DIV,l,r));
	case Absyn.OpExp.EQ: return new RelCx(Tree.CJUMP.EQ,lt,rt);
	case Absyn.OpExp.NE: return new RelCx(Tree.CJUMP.NE,lt,rt);
	case Absyn.OpExp.LT: return new RelCx(Tree.CJUMP.LT,lt,rt);
	case Absyn.OpExp.LE: return new RelCx(Tree.CJUMP.LE,lt,rt);
	case Absyn.OpExp.GT: return new RelCx(Tree.CJUMP.GT,lt,rt);
	case Absyn.OpExp.GE: return new RelCx(Tree.CJUMP.GE,lt,rt);
	default: throw new Error("Translate.OpExp");
	}
    }

    public Exp RecordExp(ExpList fields) {
	return new Ex(new Tree.CONST(1));
    }

    public Exp AssignExp(Exp lvalue, Exp rvalue) {
	return new Nx(new Tree.MOVE(lvalue.unEx(),rvalue.unEx()));
    }

    public Exp IfExp(Exp test, Exp thenclause) {
	Temp.Label t = new Temp.Label();
	Temp.Label f = new Temp.Label();

	return new Nx(new Tree.SEQ(   test.unCx(t,f),
		      new Tree.SEQ(   new Tree.LABEL(t),
                      new Tree.SEQ(   thenclause.unNx(),
				      new Tree.LABEL(f)     ))));
    }

    public Exp IfExp(Exp test, Exp thenclause, Exp elseclause) {
	Temp.Label t = new Temp.Label();
	Temp.Label f = new Temp.Label();
	Temp.Label end = new Temp.Label();

	if (thenclause instanceof Nx)
	    return new Nx(new Tree.SEQ(    test.unCx(t,f),
                          new Tree.SEQ(    new Tree.LABEL(t),
		          new Tree.SEQ(    thenclause.unNx(),
			  new Tree.SEQ(    new Tree.JUMP(end),
			  new Tree.SEQ(    new Tree.LABEL(f),
			  new Tree.SEQ(    elseclause.unNx(),
                                           new Tree.LABEL(end)        )))))));
	else {
	    Tree.TEMP r = new Tree.TEMP(new Temp.Temp());

	    return new Ex(new Tree.ESEQ(
			  new Tree.SEQ(  test.unCx(t,f),
                          new Tree.SEQ(  new Tree.LABEL(t),
                          new Tree.SEQ(  new Tree.MOVE(r,thenclause.unEx()),
                          new Tree.SEQ(  new Tree.JUMP(end),
                          new Tree.SEQ(  new Tree.LABEL(f),
                          new Tree.SEQ(  new Tree.MOVE(r,elseclause.unEx()),
			  new Tree.LABEL(end)       )))))),r));
	}
    }

    public Exp WhileExp(Exp test, Exp body, Temp.Label _break) {
	Temp.Label start = new Temp.Label();
	Temp.Label t = new Temp.Label();

	return new Nx(new Tree.SEQ(    new Tree.LABEL(start),
                      new Tree.SEQ(    test.unCx(t,_break),
                      new Tree.SEQ(    new Tree.LABEL(t),
                      new Tree.SEQ(    body.unNx(),
                      new Tree.SEQ(    new Tree.JUMP(start),
                                       new Tree.LABEL(_break)))))));
    }

    public Exp ForExp(Exp var, Exp hi, Exp body, Access v, Temp.Label _break) {
	return Nop();
    }

    public Exp BreakExp(Temp.Label l) {
	return new Nx(new Tree.JUMP(l));
    }

    public Exp LetExp(Exp decs, Exp body) {
	return Append(decs,body);
    }

    public Exp ArrayExp(Exp size, Exp init, Level l) {
	Temp.Temp s = new Temp.Temp(); Temp.Temp i = new Temp.Temp();
	Temp.Temp b = new Temp.Temp(); Temp.Temp bb = new Temp.Temp();
	int k = l.frame.wordSize();

	Temp.Label l1 = new Temp.Label(); Temp.Label l2 = new Temp.Label();
	Temp.Label l3 = new Temp.Label(); Temp.Label l4 = new Temp.Label();
	Temp.Label l5 = new Temp.Label();

	return new Ex(new Tree.ESEQ
	    (new Tree.SEQ(  new Tree.MOVE(new Tree.TEMP(s),size.unEx()),
 	     new Tree.SEQ(  new Tree.MOVE(new Tree.TEMP(i),init.unEx()),
 	     new Tree.SEQ(  new Tree.CJUMP(Tree.CJUMP.LT,new Tree.TEMP(s),
					   new Tree.CONST(0),error,l1),
	     new Tree.SEQ(  new Tree.LABEL(l1),
	     new Tree.SEQ(  new Tree.MOVE(new Tree.TEMP(b),new Tree.CALL(
                                    new Tree.NAME(malloc),
                                    new Tree.ExpList(new Tree.TEMP(s),null))),
	     new Tree.SEQ(  new Tree.CJUMP(Tree.CJUMP.EQ,new Tree.TEMP(b),
					   new Tree.CONST(0),error,l2),
             new Tree.SEQ(  new Tree.LABEL(l2),
	     new Tree.SEQ(  new Tree.MOVE(new Tree.TEMP(bb),new Tree.TEMP(b)),
	     new Tree.SEQ(  new Tree.LABEL(l3),
	     new Tree.SEQ(  new Tree.CJUMP(Tree.CJUMP.LE,new Tree.TEMP(s),
					   new Tree.CONST(0),l5,l4),
	     new Tree.SEQ(  new Tree.LABEL(l4),
	     new Tree.SEQ(  new Tree.MOVE(new Tree.MEM(new Tree.TEMP(bb)),
					  new Tree.TEMP(i)),
	     new Tree.SEQ(  new Tree.MOVE(new Tree.TEMP(bb),
		  new Tree.BINOP(Tree.BINOP.PLUS,new Tree.TEMP(bb),
				 new Tree.CONST(k))),
	     new Tree.SEQ(  new Tree.MOVE(new Tree.TEMP(s),
		  new Tree.BINOP(Tree.BINOP.MINUS,new Tree.TEMP(s),
				 new Tree.CONST(1))),
	     new Tree.SEQ(  new Tree.JUMP(l3),
                            new Tree.LABEL(l5)  ))))))))))))))),
	     new Tree.TEMP(b)));
    }

    public Exp Append(Exp head, Exp tail) {
	if (tail instanceof Nx)
	    return new Nx(new Tree.SEQ(head.unNx(),tail.unNx()));
	else
	    return new Ex(new Tree.ESEQ(head.unNx(),tail.unEx()));
    }

    public Exp Nop() {
	return new Nx(new Tree.EXP(new Tree.CONST(0)));
    }

    public Exp FunctionDec(Temp.Label name, Exp body, Level l) {
	procEntryExit(l,body);

	return Nop();
    }

    public Exp VarDec(Access v,Exp init,Level l) {
	return new Nx(new Tree.MOVE(v.acc.exp(new Tree.TEMP(l.frame.FP())),
				    init.unEx()));
    }

    public Exp TypeDec() {
	return Nop();
    }
}
